"""
This module handles the k2-connect PAY service. It entails the creation
of payments and pay recipients within the pay service. Once a transaction
is created, the service provides the user with a means to query k2 servers
for the pay transaction's status.
"""
from k2connect import service, json_builder, exceptions, validation

# PAY Service paths
ADD_PAY_PATH = "api/v1/pay_recipients"
SEND_PAY_PATH = "api/v1/payments"

# PAY recipient types
BANK_ACCOUNT_RECIPIENT_TYPE = "bank_account"
MOBILE_WALLET_RECIPIENT_TYPE = "mobile_wallet"
TILL_RECIPIENT_TYPE = "till"
PAYBILL_RECIPIENT_TYPE = "paybill"


class PayService(service.Service):
    """
    The PayService class containing methods to:
    To create pay recipients
    Example:
        >>>import k2connect
        >>>k2connect.initialize('sample_client_id', 'sample_client_secret', 'https://some_url.com/')
        >>> pay = k2-connect.Pay
        >>> pay_recipient = pay.add_pay_recipient(recipient_type='mobile_wallet',
        >>>.......................................first_name='Jaqen',
        >>>.......................................last_name='Hghar',
        >>>.......................................phone='+007856798451',
        >>>.......................................network='Valyria Mobile')
    To send payments to third parties
    Example:
        >>> pay_transaction = pay.send_pay('https://mycoolapp.com',
        >>>................................'c7f300c0-f1ef-4151-9bbe-005005aa3747',
        >>>................................'25000',
        >>>................................customerId='8675309',
        >>>................................notes='Salary payment for May 2018')
    To check transaction statuses
    Example:
        >>> pay.transaction status(pay_transaction)
        '{"status":"Scheduled","reference":"KKKKKKKKK",
        "origination_time":"2018-07-20T22:45:12.790Z",
        "destination":"c7f300c0-f1ef-4151-9bbe-005005aa3747",
        "amount":{"currency":"KES","value":20000},"
        metadata":{"customerId":"8675309",
        "notes":"Salary payment for May 2018"},
        "_links":{"self":"https://api-sandbox.kopokopo.com/payments
        /d76265cd-0951-e511-80da-0aa34a9b2388"}}'
    get payment request location
    Example:object
        >>> pay.pay_transaction_location(pay_transaction)
        'https://api-sandbox.kopokopo.com/payments/c7f300c0-f1ef-4151-9bbe-005005aa3747'
    """

    def __init__(self, base_url):
        """
        :param base_url: The domain to use in the library.
        :type base_url: str
        """
        super(PayService, self).__init__(base_url=base_url)

    def add_pay_recipient(self, kwargs):
        """
        Adds external entities that will be the destination of your payments.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return:'requests.models.Response'
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')
        if 'recipient_type' not in kwargs:
            raise exceptions.InvalidArgumentError('Recipient Type not given.')
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'recipient_type' in kwargs:
            recipient_type = kwargs['recipient_type']

        # define headers
        headers = dict(self._headers)

        validation.validate_string_arguments(bearer_token)

        # add bearer token
        headers['Authorization'] = 'Bearer ' + bearer_token

        # build url
        add_pay_url = self._build_url(url_path=ADD_PAY_PATH)

        if 'email' in kwargs:
            validation.validate_email(str(kwargs['email']))
        if 'phone_number' in kwargs:
            validation.validate_phone_number(str(kwargs['phone_number']))

        # expected parameters for bank account wallet recipient
        if recipient_type == BANK_ACCOUNT_RECIPIENT_TYPE:
            if 'account_name' not in kwargs or \
                    'account_number' not in kwargs or \
                    'settlement_method' not in kwargs or \
                    'bank_branch_ref' not in kwargs:
                raise exceptions.InvalidArgumentError('Invalid arguments for bank account Pay recipient')

            # build recipient json object
            recipient_object = json_builder.bank_account(account_name=str(kwargs['account_name']),
                                                         account_number=str(kwargs['account_number']),
                                                         settlement_method=str(kwargs['settlement_method']),
                                                         bank_branch_ref=str(kwargs['bank_branch_ref']))
            # build bank payment recipient json object
            payment_recipient_object = json_builder.pay_recipient(recipient_type=recipient_type,
                                                                  recipient=recipient_object)
        # expected parameters for mobile wallet recipient
        # ['first_name', 'last_name',
        # network','phone_number','email']

        elif recipient_type == MOBILE_WALLET_RECIPIENT_TYPE:
            if 'first_name' not in kwargs or \
                    'last_name' not in kwargs or \
                    'phone_number' not in kwargs or \
                    'email' not in kwargs or \
                    'network' not in kwargs:
                raise exceptions.InvalidArgumentError('Invalid arguments for mobile wallet Pay recipient')

            # create recipient json object
            recipient_object = json_builder.mobile_wallet(first_name=str(kwargs['first_name']),
                                                          last_name=str(kwargs['last_name']),
                                                          phone_number=str(kwargs['phone_number']),
                                                          network=str(kwargs['network']),
                                                          email=str(kwargs['email']))

            # create mobile wallet recipient json object
            payment_recipient_object = json_builder.pay_recipient(recipient_type=recipient_type,
                                                                  recipient=recipient_object)

        elif recipient_type == TILL_RECIPIENT_TYPE:
            if 'till_name' not in kwargs or \
                    'till_number' not in kwargs:
                raise exceptions.InvalidArgumentError('Invalid arguments for till Pay recipient')

            # create recipient json object
            recipient_object = json_builder.till_pay_recipient(till_name=str(kwargs['till_name']),
                                                          till_number=str(kwargs['till_number']))

            # create mobile wallet recipient json object
            payment_recipient_object = json_builder.pay_recipient(recipient_type=recipient_type,
                                                                  recipient=recipient_object)
        # expected parameters for mobile wallet recipient
        # ['till_name', 'till_number']

        elif recipient_type == PAYBILL_RECIPIENT_TYPE:
            if 'paybill_name' not in kwargs or \
                    'paybill_number' not in kwargs or \
                    'paybill_account_number' not in kwargs:
                raise exceptions.InvalidArgumentError('Invalid arguments for paybill Pay recipient')

            # create recipient json object
            recipient_object = json_builder.paybill_pay_recipient(paybill_name=str(kwargs['paybill_name']),
                                                                  paybill_number=str(kwargs['paybill_number']),
                                                                  paybill_account_number=str(kwargs['paybill_account_number']))

            # create mobile wallet recipient json object
            payment_recipient_object = json_builder.pay_recipient(recipient_type=recipient_type,
                                                                  recipient=recipient_object)
        # expected parameters for mobile wallet recipient
        # ['alias_name', 'till_number']

        else:
            raise exceptions.InvalidArgumentError('The recipient type is not recognized by k2connect')

        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=add_pay_url,
                                   payload=payment_recipient_object)

    def send_pay(self, kwargs):
        """
        Creates an outgoing pay to a third party. The result of
        the pay is provided asynchronously and posted to the callback_url
        provided.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: Provision for optional metadata with maximum of 5
        key value pairs.
        :type kwargs: dict

        :return:requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'destination_reference' not in kwargs or \
                'destination_type' not in kwargs or \
                'callback_url' not in kwargs or \
                'description' not in kwargs or \
                'amount' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Outgoing Pay.')

        if 'currency' not in kwargs:
            currency = 'KES'

        if 'metadata' not in kwargs:
            pay_metadata = ''

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'callback_url' in kwargs:
            callback_url = kwargs['callback_url']
        if 'description' in kwargs:
            description = kwargs['description']
        if 'currency' in kwargs:
            currency = 'KES'
        if 'metadata' in kwargs:
            pay_metadata = json_builder.metadata(kwargs['metadata'])

        # build send_pay url
        send_pay_url = self._build_url(SEND_PAY_PATH)

        # define headers
        headers = dict(self._headers)

        # check bearer token
        validation.validate_string_arguments(bearer_token)

        # add authorization to headers
        headers['Authorization'] = 'Bearer ' + bearer_token + ''

        # create amount json object
        pay_amount = json_builder.amount(currency=currency,
                                         value=kwargs['amount'])

        # create links json object
        pay_links = json_builder.links(callback_url=callback_url)

        # create payment json object
        pay_json = json_builder.pay(kwargs['destination_reference'],
                                    kwargs['destination_type'],
                                    pay_amount,
                                    description,
                                    pay_links,
                                    pay_metadata)

        return self._make_requests(url=send_pay_url,
                                   method='POST',
                                   payload=pay_json,
                                   headers=headers)

    def pay_transaction_status(self,
                               bearer_token,
                               query_url):
        """
        Returns a JSON object result containing the transaction status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)

    @staticmethod
    def pay_transaction_location(response):
        """
        Returns location of the pay transaction result as returned in the headers of the
        response body.
        :param response: response object from a HTTP request
        :type response: requests.models.Response
        :return str
        """
        return service.k2_requests.get_location(response)



"""

This module handles k2-connect receive MPESA payments service. It facilitates
the reception of payments from MPESA users. It creates requests for the reception
of MPESA payments. Succeeding a MPESA payment request, it queries the status of the
transaction.
"""
from k2connect import exceptions
from k2connect import json_builder
from k2connect import k2_requests
from k2connect import service
from k2connect import validation

POLLING_PATH = 'api/v1/polling'


class PollingService(service.Service):
    """
    The ReceivePaymentsService class containing methods to create requests for
    MPESA payments:
    Example:
        >>>import k2connect
        >>>k2connect.initialize('sample_client_id', 'sample_client_secret', 'https://some_url.com/')
        >>>polling = k2connect.PollingService
        >>>polling_request = polling.create_polling_request('bearer_token',
        >>>....................................................................'callback_url',
        >>>....................................................................'currency',
        >>>....................................................................'first_name',
        >>>....................................................................'last_name',
        >>>....................................................................'payment_channel')

    To check payment request status:
    Example:
        >>>polling_status = polling.polling_request_status('bearer_token',
        >>>......................................................'query_url')

    To check payment request location:
    Example:
        >>>polling_location = polling.polling_request_location('response')
    """

    def __init__(self, base_url):
        """
        :param base_url: The domain to use in the library
        :type  base_url: str
        """
        super(PollingService, self).__init__(base_url=base_url)

    # noinspection PyArgumentList
    def create_polling_request(self, kwargs):
        """
        Creates a request for the reception of payments from MPESA users.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'scope' not in kwargs or \
                'scope_reference' not in kwargs or \
                'from_time' not in kwargs or \
                'to_time' not in kwargs or \
                'callback_url' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Polling Request.')

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'scope' in kwargs:
            scope = kwargs['scope']
        if 'scope_reference' in kwargs:
            scope_reference = kwargs['scope_reference']
        if 'from_time' in kwargs:
            from_time = kwargs['from_time']
        if 'to_time' in kwargs:
            to_time = kwargs['to_time']
        if 'callback_url' in kwargs:
            callback_url = kwargs['callback_url']

        # define headers
        headers = dict(self._headers)

        # validate bearer_token
        validation.validate_string_arguments(bearer_token)

        # add bearer token
        headers['Authorization'] = 'Bearer ' + bearer_token + ''

        # build polling request url
        polling_request_url = self._build_url(POLLING_PATH)

        # define links JSON object
        polling_request_links = json_builder.links(callback_url=callback_url)

        # define MPESA payment request JSON object
        polling_request_payload = json_builder.polling(scope=scope,
                                                       scope_reference=scope_reference,
                                                       from_time=from_time,
                                                       to_time=to_time,
                                                       polling_links=polling_request_links)
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=polling_request_url,
                                   payload=polling_request_payload)

    def polling_request_status(self,
                               bearer_token,
                               query_url):
        """
        Returns a JSON object result containing the payment request status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)

    @staticmethod
    def polling_request_location(response):
        """
        Returns location of the receive mpesa transaction result as returned in the headers of the
        response body.
        :param response: response object from a HTTP request
        :type response: requests.models.Response
        :return str
        """
        return service.k2_requests.get_location(response)



"""Handles validation of expected user inputs"""
import re
from urllib.parse import urlparse
from k2connect import exceptions


def validate_email(email):
    """
    :param email: An email address
    :type email: str
    :return: True
    """
    validated_email = re.search(r'\b[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,}\b', email, re.I)
    if validated_email is not None:
        return True
    else:
        raise exceptions.InvalidArgumentError("The email address passed is invalid. (Should be: email@domain.com)")


def validate_phone_number(phone_number):
    """
    :param phone_number: A phone number
    :type phone_number: str
    :return: True
    """
    validated_phone_number = re.search('^\+\d{1,3}\d{3,}$', phone_number)
    if validated_phone_number is not None:
        return True
    else:
        raise exceptions.InvalidArgumentError('The phone number passed is invalid. (Should be: +254123456789)')


def validate_url(url):
    validated_url = urlparse(url)
    # check url format
    if validated_url.scheme is "" or validated_url.netloc is "":
        raise exceptions.InvalidArgumentError('The url format passed is invalid (should be : https://domain.com)')
    if validated_url.scheme != "http" and validated_url.scheme != "https":
        raise exceptions.InvalidArgumentError('Provide a url with a valid certificate => (http://) or (https://')
    else:
        return True


def validate_base_url(url):
    validate_url(url)
    validated_url = urlparse(url)
    # check url format
    if validated_url.path is not "/":
        raise exceptions.InvalidArgumentError('Invalid base url format (should be : https://domain.com/ with the domain'
                                              ' details only)')
    else:
        return True


def validate_string_arguments(*args):
    for arg in args:
        if arg is None or arg == '':
            raise exceptions.InvalidArgumentError('Invalid/empty argument')
        if not isinstance(arg, str):
            raise exceptions.InvalidArgumentError('Argument must be of type str.')
        pass


def validate_dictionary_arguments(*args):
    for arg in args:
        if arg is None or arg == {}:
            raise exceptions.InvalidArgumentError('Invalid argument passed.'
                                                  'Expects dict object')
        if not isinstance(arg, dict):
            raise exceptions.InvalidArgumentError('Argument must be of type dict.')



"""
This module is responsible for the processing of payloads in k2-connect.
It converts JSON objects to python dictionaries then breaks down the python
dictionaries and sets their values to predefined "names" that correspond to
their respective keys in the JSON object using python properties.
:Example:
    sample_dictionary = {'key': 'value'}
    instance_of_payload_decomposer.key = sample_dictionary['value']
"""
import json
from k2connect import exceptions
from k2connect import k2_model
from k2connect import validation

B2B_TRANSACTION_RECEIVED = 'b2b_transaction_received'
BUYGOODS_TRANSACTION_RECEIVED = 'buygoods_transaction_received'
CREATE_PAYMENT = 'payment'
RECEIVE_PAYMENTS = 'incoming_payment'
TRANSFER = 'settlement_transfer'
# TODO: Pending
BUYGOODS_TRANSACTION_REVERSED = 'buygoods_transaction_reversed'
CUSTOMER_CREATED = 'customer_created'
MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED = 'm2m_transaction_received'
SETTLEMENT_TRANSFER_COMPLETED = 'settlement_transfer_completed'


def decompose(json_payload):
    """
    Returns an instance of the K2Model class. The decomposer object is used to
    set values that are then accessible through python properties of the same name
    as the setters.
    Example:
        For the JSON object:
        x = {
             "resource": {
                "sender_first_name": "Tyrion",
                "sender_last_name": "Lanister"
                }
            }


        >>>from k2connect import k2_model
        >>>decomposer = k2_model.K2Model()
        >>>resource_payload_nest = payload_dictionary['event']['resource']
        >>>decomposer.first_name = resource_payload_nest['sender_first_name']
        test
        >>> print(decomposer.first_name)
        'Tyrion'

    :param json_payload: A JSON formatted str object
    :type json_payload: str
    """
    validation.validate_string_arguments(*json_payload)

    # convert json object to python dictionary for decomposing
    payload_dictionary = json.loads(json_payload)

    # validate dictionary
    validation.validate_dictionary_arguments(payload_dictionary)

    # define decomposer
    decomposer = k2_model.K2Model()

    # Webhooks
    if 'data' not in payload_dictionary:
        result_topic = payload_dictionary['topic']
        decomposer.topic = result_topic
        webhook_event_nest = payload_dictionary['event']
        webhook_result_type = webhook_event_nest['type']
        resource_payload_nest = webhook_event_nest['resource']
        links_payload_nest = payload_dictionary['_links']
        decomposer.result_type = webhook_result_type
        webhook_decompose(decomposer, result_topic, payload_dictionary, resource_payload_nest, links_payload_nest)
    elif 'data' in payload_dictionary:
        # Payments
        data_payload_nest = payload_dictionary['data']
        payments_result_type = data_payload_nest['type']
        payments_attributes_payload_nest = data_payload_nest['attributes']
        payments_metadata_payload_nest = payments_attributes_payload_nest['metadata']
        payments_links_payload_nest = payments_attributes_payload_nest['_links']
        payment_decompose(decomposer, data_payload_nest, payments_result_type, payments_attributes_payload_nest,
                          payments_links_payload_nest)
        decomposer.metadata = payments_metadata_payload_nest
    return decomposer


def webhook_decompose(decomposer, result_topic, payload_dictionary, resource_payload_nest, links_payload_nest):
    # decompose ID values that are similar in ALL WEBHOOKS
    if result_topic == B2B_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_REVERSED \
            or result_topic == CUSTOMER_CREATED \
            or result_topic == MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED \
            or result_topic == SETTLEMENT_TRANSFER_COMPLETED:
        decomposer.id = payload_dictionary['id']
        decomposer.created_at = payload_dictionary['created_at']

    if result_topic == B2B_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_REVERSED \
            or result_topic == CUSTOMER_CREATED \
            or result_topic == MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED:
        decomposer.resourceId = resource_payload_nest['id']

    # decompose status value that are similar in TRANSACTIONS
    if result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == B2B_TRANSACTION_RECEIVED \
            or result_topic == MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_REVERSED \
            or result_topic == SETTLEMENT_TRANSFER_COMPLETED:
        decomposer.status = resource_payload_nest['status']
        decomposer.currency = resource_payload_nest['currency']
        decomposer.amount = resource_payload_nest['amount']
        decomposer.reference = resource_payload_nest['reference']
        decomposer.origination_time = resource_payload_nest['origination_time']

    # decompose system value that are similar in TRANSACTIONS
    if result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == B2B_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_REVERSED:
        decomposer.system = resource_payload_nest['system']
        decomposer.till_number = resource_payload_nest['till_number']

    # decompose all values that are similar in BUYGOODS TRANSACTIONS
    if result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == BUYGOODS_TRANSACTION_REVERSED:
        decomposer.first_name = resource_payload_nest['sender_first_name']
        # decomposer.middle_name = resource_payload_nest['sender_middle_name']
        decomposer.last_name = resource_payload_nest['sender_last_name']

    # decompose all values that have similar links structures
    if result_topic == BUYGOODS_TRANSACTION_REVERSED \
            or result_topic == BUYGOODS_TRANSACTION_RECEIVED \
            or result_topic == SETTLEMENT_TRANSFER_COMPLETED \
            or result_topic == CUSTOMER_CREATED \
            or result_topic == B2B_TRANSACTION_RECEIVED \
            or result_topic == MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED:
        decomposer.links_self = links_payload_nest['self']
        decomposer.links_resource = links_payload_nest['resource']

        # decompose value specific to buygoods transaction reversed
        if result_topic == BUYGOODS_TRANSACTION_REVERSED:
            decomposer.reversal_time = resource_payload_nest['reversal_time']

    # decompose unique to b2b transaction received
    if result_topic == B2B_TRANSACTION_RECEIVED:
        decomposer.sending_till = resource_payload_nest['sending_till']
        decomposer.till_number = resource_payload_nest['till_number']

    # decompose unique to merchant to merchant transaction received
    if result_topic == MERCHANT_TO_MERCHANT_TRANSACTION_RECEIVED:
        decomposer.sending_merchant = resource_payload_nest['sending_merchant']

    # decompose unique to customer created
    if result_topic == CUSTOMER_CREATED:
        decomposer.first_name = resource_payload_nest['first_name']
        decomposer.middle_name = resource_payload_nest['middle_name']
        decomposer.last_name = resource_payload_nest['last_name']

    # decompose unique to settlement transfer completed
    if result_topic == SETTLEMENT_TRANSFER_COMPLETED:
        destination_payload_nest = resource_payload_nest['destination']
        decomposer.destination_type = destination_payload_nest['type']
        destination_resource_payload_nest = destination_payload_nest['resource']
        if decomposer.destination_type == 'merchant_bank_account':
            decomposer.bank_id = destination_resource_payload_nest['bank_id']
            decomposer.bank_branch_id = destination_resource_payload_nest['bank_branch_id']
            decomposer.account_name = destination_resource_payload_nest['account_name']
            decomposer.account_number = destination_resource_payload_nest['account_number']
        elif decomposer.destination_type == 'merchant_wallet':
            decomposer.msisdn = destination_resource_payload_nest['msisdn']
            decomposer.network = destination_resource_payload_nest['network']


def payment_decompose(decomposer, data_payload_nest, payments_result_type, payments_attributes_payload_nest,
                      payments_links_payload_nest):
    # decompose all values that are common to PAYMENTS(OUTGOING OR INCOMING) and Transfer
    if payments_result_type == RECEIVE_PAYMENTS \
            or payments_result_type == CREATE_PAYMENT \
            or payments_result_type == TRANSFER:
        decomposer.id = data_payload_nest['id']
        decomposer.type = payments_result_type
        decomposer.status = payments_attributes_payload_nest['status']
        decomposer.self = payments_links_payload_nest['self']
        decomposer.callback_url = payments_links_payload_nest['callback_url']

    # decompose all values that are common to receive MPESA payments service (INCOMING)
    if payments_result_type == RECEIVE_PAYMENTS:
        decomposer.initiation_time = payments_attributes_payload_nest['initiation_time']
        # Event Payload
        payments_events_payload_nest = payments_attributes_payload_nest['event']
        decomposer.errors = payments_events_payload_nest['errors']
        decomposer.event_type = payments_events_payload_nest['type']
        # incoming payments resource payload
        payments_resource_payload_nest = payments_events_payload_nest['resource']
        decomposer.resource_id = payments_resource_payload_nest['id']
        decomposer.transaction_reference = payments_resource_payload_nest['reference']
        decomposer.origination_time = payments_resource_payload_nest['origination_time']
        decomposer.sender_phone_number = payments_resource_payload_nest['sender_phone_number']
        decomposer.amount = payments_resource_payload_nest['amount']
        decomposer.currency = payments_resource_payload_nest['currency']
        decomposer.till_number = payments_resource_payload_nest['till_number']
        decomposer.system = payments_resource_payload_nest['system']
        decomposer.resource_status = payments_resource_payload_nest['status']
        decomposer.sender_last_name = payments_resource_payload_nest['sender_last_name']
        decomposer.sender_first_name = payments_resource_payload_nest['sender_first_name']
        decomposer.sender_middle_name = payments_resource_payload_nest['sender_middle_name']

    # decompose all values that are common to the PAY (OUTGOING PAYMENTS) and TRANSFER service
    if payments_result_type == CREATE_PAYMENT \
            or payments_result_type == TRANSFER:
        decomposer.created_at = payments_attributes_payload_nest['created_at']
        decomposer.origination_time = payments_attributes_payload_nest['origination_time']
        decomposer.amount = payments_attributes_payload_nest['amount']['value']
        decomposer.currency = payments_attributes_payload_nest['amount']['currency']
        decomposer.transfer_batches = payments_attributes_payload_nest['transfer_batches']

    # decompose all values that are in the PAY service (OUTGOING PAYMENTS)
    if payments_result_type == CREATE_PAYMENT:
        decomposer.metadata = payments_attributes_payload_nest['metadata']



"""
This module handles the initialization of the k2connect library.
It takes a client id, secret and base url and initializes all k2connect
services as with appropriate required arguments.
"""
# TODO: Remember to remove http from validation
# TODO: David-dev branch is the one that is behind use the development branch which is the updated one
from k2connect import exceptions

import k2connect

from .authorization import TokenService
from .pay import PayService
from .result_processor import ResultProcessor
from .receive_payments import ReceivePaymentsService
from .transfers import TransferService
from . import validation
from .webhooks import WebhookService
from .notifications import NotificationService
from .polling import PollingService


Tokens = None
ReceivePayments = None
Pay = None
Transfers = None
Webhooks = None
TransactionNotifications = None
ResultHandler = None
Polling = None
__version__ = '1.2.0'


def initialize(client_id, client_secret, base_url, api_secret=None):
    """
    Initializes k2connect services
    :param base_url: The domain to use in the library.
    :type base_url: str
    :param client_id: Identifier for the k2 user.
    :type client_id: str
    :param client_secret: Secret key for k2 user.
    :type client_secret: str
    :param api_secret: API Secret key for processing webhook payloads.
    :type api_secret: str
    """
    validation.validate_string_arguments(client_id,
                                         client_secret,
                                         base_url)
    validation.validate_base_url(base_url)

    # initialize  token service
    globals()['Tokens'] = TokenService(client_id=client_id,
                                       client_secret=client_secret,
                                       base_url=base_url)

    # initialize stk service
    globals()['ReceivePayments'] = ReceivePaymentsService(base_url=base_url)

    # initialize Pay service
    globals()['Pay'] = PayService(base_url=base_url)

    # initialize transfers service
    globals()['Transfers'] = TransferService(base_url=base_url)

    # initialize webhook service
    globals()['Webhooks'] = WebhookService(base_url=base_url)

    # initialize transaction notification service
    globals()['TransactionNotifications'] = NotificationService(base_url=base_url)

    # initialize polling service
    globals()['Polling'] = PollingService(base_url=base_url)

    # initialize response processor
    globals()['ResultHandler'] = ResultProcessor(base_url=base_url,
                                                 api_secret=api_secret)



"""
This module handles transfer of funds to pre-approved settlement accounts. It creates
verified settlement bank accounts. Once an account to which funds can be settled is available,
it creates blind and targeted settlement transactions.
Blind transfers are made to the default settlement account defined upon k2 customer acquisition,
targeted transfers are made to a defined destination account.
"""
from k2connect import json_builder, exceptions, service, validation

TRANSACTION_NOTIFICATIONS_PATH = 'api/v1/transaction_sms_notifications'


class NotificationService(service.Service):
    """
    The TransferService class containing methods for creation of settlement accounts:
    Example:
        # initialize transfer service
        >>> k2connect.TransferService
        >>> k2connect.add_bank_settlement_account('Arya Stark',
        >>>.....................................'45164-IRON BANK',
        >>>.....................................'78491631254523',
        >>>.....................................'564123456987845')
    """

    def __init__(self, base_url):
        """
        Initializes transfer services with the base_url as an argument.
        :param base_url:
        :type  base_url: str
        """
        super(NotificationService, self).__init__(base_url)

    def send_transaction_sms_notification(self, kwargs):
        """
        Creates a transaction sms notification.
        Returns a request response object < class, 'requests.models.Response'>
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'webhook_event_reference' not in kwargs or \
                'message' not in kwargs or \
                'callback_url' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Transaction SMS Notification.')

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'webhook_event_reference' in kwargs:
            webhook_event_reference = kwargs['webhook_event_reference']
        if 'message' in kwargs:
            message = kwargs['message']
        if 'callback_url' in kwargs:
            callback_url = kwargs['callback_url']

        # build url
        create_transaction_sms_notification_url = self._build_url(TRANSACTION_NOTIFICATIONS_PATH)

        # define headers
        headers = dict(self._headers)

        # validate string arguments
        validation.validate_string_arguments(bearer_token, webhook_event_reference, message, callback_url)

        notification_links = json_builder.links(callback_url=callback_url)

        # add authorization to headers
        headers['Authorization'] = 'Bearer ' + bearer_token + ''
        # define create bank settlement account payload
        create_transaction_sms_notification_payload = json_builder.transaction_sms_notification(webhook_event_reference=webhook_event_reference,
                                                                                                message=message,
                                                                                                notification_links=notification_links)
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=create_transaction_sms_notification_url,
                                   payload=create_transaction_sms_notification_payload)

    def transaction_notification_status(self,
                                    bearer_token,
                                    query_url):
        """
        Returns a JSON object result containing the transaction status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API.
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)

    @staticmethod
    def transaction_notification_location(response):
        """
        Returns location of the transfers transaction result as returned in the headers of the
        response body.
        :param response: response object from a HTTP request
        :type response: requests.models.Response
        :return str
        """
        return service.k2_requests.get_location(response)



import setuptools

with open("README.md", "r") as file_header:
    long_description = file_header.read()

setuptools.setup(
    name='k2-connect',
    version='1.2.2',
    author='Philip Wafula, David Mwangi',
    author_email='philipwafula2@gmail.com, david.mwangi@kopokopo.com',
    description='A python SDK to connect to Kopo Kopo API',
    long_description=long_description,
    python_requires='>=3',
    long_description_content_type='text/markdown',
    url='https://github.com/kopokopo/k2-connect-python',
    license='MIT',
    packages=setuptools.find_packages(exclude=['docs', 'tests*', 'example']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



"""

This module handles k2-connect receive MPESA payments service. It facilitates
the reception of payments from MPESA users. It creates requests for the reception
of MPESA payments. Succeeding a MPESA payment request, it queries the status of the
transaction.
"""
from k2connect import exceptions
from k2connect import json_builder
from k2connect import k2_requests
from k2connect import service
from k2connect import validation

CREATE_RECEIVE_MPESA_PAYMENT_PATH = 'api/v1/incoming_payments'


class ReceivePaymentsService(service.Service):
    """
    The ReceivePaymentsService class containing methods to create requests for
    MPESA payments:
    Example:
        >>>import k2connect
        >>>k2connect.initialize('sample_client_id', 'sample_client_secret', 'https://some_url.com/')
        >>>mpesa_payment = k2connect.ReceiveMpesaPayments
        >>>mpesa_payment_request = mpesa_payment.create_payment_request('bearer_token',
        >>>....................................................................'callback_url',
        >>>....................................................................'currency',
        >>>....................................................................'first_name',
        >>>....................................................................'last_name',
        >>>....................................................................'payment_channel',
        >>>....................................................................'phone',
        >>>....................................................................'till_number',
        >>>....................................................................'value')

    To check payment request status:
    Example:
        >>>payment_status = mpesa_payment.payment_request_status('bearer_token',
        >>>......................................................'query_url')

    To check payment request location:
    Example:
        >>>payment_location = mpesa_payment.payment_request_location('response')
    """

    def __init__(self, base_url):
        """
        :param base_url: The domain to use in the library
        :type  base_url: str
        """
        super(ReceivePaymentsService, self).__init__(base_url=base_url)

    # noinspection PyArgumentList
    def create_payment_request(self, kwargs):
        """
        Creates a request for the reception of payments from MPESA users.
        Returns a request response object < class, 'requests.models.Response'>
        :param kwargs: The values constitute all user input.
        :type kwargs: dict
        :return: requests.models.Response
        """
        if 'access_token' not in kwargs:
            raise exceptions.InvalidArgumentError('Access Token not given.')

        if 'first_name' not in kwargs or \
                'last_name' not in kwargs or \
                'callback_url' not in kwargs or \
                'payment_channel' not in kwargs or \
                'phone_number' not in kwargs or \
                'till_number' not in kwargs or \
                'email' not in kwargs or \
                'amount' not in kwargs:
            raise exceptions.InvalidArgumentError('Invalid arguments for creating Incoming Payment Request.')

        if 'currency' not in kwargs:
            currency = 'KES'

        if 'metadata' not in kwargs:
            mpesa_payment_metadata = ''

        # iterate through kwargs
        if 'access_token' in kwargs:
            bearer_token = kwargs['access_token']
        if 'phone_number' in kwargs:
            phone_number = kwargs['phone_number']
            if validation.validate_phone_number(phone_number) is False:
                pass
        if 'email' in kwargs:
            email = kwargs['email']
            validation.validate_email(email)
        if 'currency' in kwargs:
            currency = 'KES'
        if 'metadata' in kwargs:
            mpesa_payment_metadata = json_builder.metadata(kwargs['pay_metadata'])

        # define headers
        headers = dict(self._headers)

        # validate bearer_token
        validation.validate_string_arguments(bearer_token)

        # add bearer token
        headers['Authorization'] = 'Bearer ' + bearer_token + ''

        # build create mpesa payment request url
        mpesa_payment_request_url = self._build_url(CREATE_RECEIVE_MPESA_PAYMENT_PATH)

        # define amount JSON object
        mpesa_payment_request_amount = json_builder.amount(currency=currency,
                                                           value=kwargs['amount'])

        # define links JSON object
        mpesa_payment_request_links = json_builder.links(callback_url=kwargs['callback_url'])

        # define subscriber JSON object
        mpesa_payment_subscriber = json_builder.subscriber(first_name=kwargs['first_name'],
                                                           last_name=kwargs['last_name'],
                                                           phone_number=phone_number,
                                                           email=email)

        # define MPESA payment request JSON object
        mpesa_payment_request_payload = json_builder.mpesa_payment(mpesa_links=mpesa_payment_request_links,
                                                                   mpesa_payment_amount=mpesa_payment_request_amount,
                                                                   mpesa_payment_subscriber=mpesa_payment_subscriber,
                                                                   metadata=mpesa_payment_metadata,
                                                                   payment_channel=kwargs['payment_channel'],
                                                                   till_number=kwargs['till_number'])
        return self._make_requests(headers=headers,
                                   method='POST',
                                   url=mpesa_payment_request_url,
                                   payload=mpesa_payment_request_payload)

    def payment_request_status(self,
                               bearer_token,
                               query_url):
        """
        Returns a JSON object result containing the payment request status.
        :param bearer_token: Access token to be used to make calls to
        the Kopo Kopo API
        :type bearer_token: str
        :param query_url: URL to which status query is made.
        :type query_url: str
        :return str
        """
        return self._query_transaction_status(bearer_token=bearer_token,
                                              query_url=query_url)

    @staticmethod
    def payment_request_location(response):
        """
        Returns location of the receive mpesa transaction result as returned in the headers of the
        response body.
        :param response: response object from a HTTP request
        :type response: requests.models.Response
        :return str
        """
        return service.k2_requests.get_location(response)
